# v1.11, 2 August  2017
# author: Brian Fiedler, University of Oklahoma
# MIT license.
# 
import numpy as np

###########################
# make stretched x and y grids
def makexy(xres,yres,xmax,ymax,stx,sty):
    g={}
    gx=np.linspace(0,1.,xres)
    x1=xmax*(stx*gx+(2.-stx)*gx**3)/(1.+gx**2)
    gy=np.linspace(0,1.,yres)
    y1=ymax*(sty*gy+(2.-sty)*gy**3)/(1.+gy**2)
    x,y=np.meshgrid(x1,y1)
    return x,y

###########################
# make finite-difference coefficients for grid 
def makegridcoef(x,y):
    g={}
    g['x'],g['y']=x,y

    xc=(x[:,:-1]+x[:,1:])/2.
    xa2=x**2
    xc2=xc**2
    xc3=xc**3
    g['xc'], g['xa2'] ,g['xc2'], g['xc3'] = xc,xa2,xc2,xc3

    rdxc=1./(x[:,1:]-x[:,:-1])
    rdyc=1./(y[1:,:]-y[:-1,:])
    rdxa=0.*x
    rdya=0.*x
    rdxa[:,1:-1]=2./(x[:,2:]-x[:,:-2])
    rdxa[:,0]=rdxc[:,0]
    rdxa[:,-1]=rdxc[:,-1]
    rdya[1:-1,:]=2./(y[2:,:]-y[:-2,:])
    rdya[0,:]=rdyc[0,:]
    rdya[-1,:]=rdyc[-1,:]
    g['rdxc'], g['rdyc'], g['rdxa'], g['rdya'] = rdxc, rdyc, rdxa, rdya 

# for use with axisymmetric simulations
    rx=1./xc[:-1,:]
    rdrr=0.*x
    rdrr[:,1:-1]=2./(xc2[:,1:]-xc2[:,:-1])
    rdrr[:,0]= 2./xc2[:,0]
    rdrr[:,-1]= 2./(x[:,-1]**2-xc2[:,-1])
    rdrra=2./(xa2[:,1:]-xa2[:,:-1])
    g['rx'], g['rdrr'], g['rdrra'] = rx, rdrr, rdrra
    return g

############################
# 3rd order advection of q
def advect3(q,u,v,grid):
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    dqdt=np.zeros(q.shape)
    dqdxm=np.zeros(q.shape)
    dqdxp=np.zeros(q.shape)
    dqdym=np.zeros(q.shape)
    dqdyp=np.zeros(q.shape)

# for horizontal direction
#  m-> extends in "minus direction", use with positive u
#  p-> extends in "positive direction", use with negative u
    dqdxm[:,1]=(q[:,1]-q[:,0])  # 1st order near left boundary  
    dqdxm[:,2:-1]=(2*q[:,3:]+3*q[:,2:-1]-6*q[:,1:-2]+q[:,:-3])/6. # 3rd order
    dqdxp[:,-2]=(q[:,-1]-q[:,-2]) # 1st order near right boundary
    dqdxp[:,1:-2]=-(2*q[:,0:-3]+3*q[:,1:-2]-6*q[:,2:-1]+q[:,3:])/6. # 3rd order

# for vertical direction
    dqdym[1,:]=(q[1,:]-q[0,:]) # 1st
    dqdym[2:-1,:]=(2*q[3:,:]+3*q[2:-1,:]-6*q[1:-2,:]+q[:-3,:])/6. # 3rd
    dqdyp[-2,:]=(q[-1,:]-q[-2,:]) # 1st
    dqdyp[1:-2,:]=-(2*q[0:-3,:]+3*q[1:-2,:]-6*q[2:-1,:]+q[3:,:])/6. # 3rd

    dqdx=np.where(u>0.,dqdxm,dqdxp)
    dqdy=np.where(v>0.,dqdym,dqdyp)

    dqdt+=-u*dqdx*rdxa
    dqdt+=-v*dqdy*rdya

    return dqdt

################################
# 5th order advection of q
def advect5(q,u,v,grid):
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    dqdt=np.zeros(q.shape)
    dqdxm=np.zeros(q.shape)
    dqdxp=np.zeros(q.shape)
    dqdym=np.zeros(q.shape)
    dqdyp=np.zeros(q.shape)

# for horizontal
#  m-> extends in "minus direction", use with positive u
#  p-> extends in "positive direction", use with negative u
    dqdxm[:,1]=(q[:,1]-q[:,0])  # 1st order near left  boundary
#    dqdxm[:,1]=(q[:,2]-q[:,0])/2. # 2nd order
    dqdxm[:,2]=(2*q[:,3]+3*q[:,2]-6*q[:,1]+q[:,0])/6. # 3rd order near left boundary
    dqdxm[:,-2]=(2*q[:,-1]+3*q[:,-2]-6*q[:,-3]+q[:,-4])/6. # 3rd order near right boundary
    dqdxm[:,3:-2]=(-3*q[:,5:]+30*q[:,4:-1]+20*q[:,3:-2]-60*q[:,2:-3]+15*q[:,1:-4]-2*q[:,:-5])/60. # 5th order

    dqdxp[:,-2]=(q[:,-1]-q[:,-2]) # 1st 
#    dqdxp[:,-2]=(q[:,-1]-q[:,-3])/2. # 2nd
    dqdxp[:,-3]=-(2*q[:,-4]+3*q[:,-3]-6*q[:,-2]+q[:,-1])/6. # 3rd
    dqdxp[:,1]=-(2*q[:,0]+3*q[:,1]-6*q[:,2]+q[:,3])/6. # 3rd
    dqdxp[:,2:-3]=-(-3*q[:,:-5]+30*q[:,1:-4]+20*q[:,2:-3]-60*q[:,3:-2]+15*q[:,4:-1]-2*q[:,5:])/60. # 5th


# for vertical 
    dqdym[1,:]=(q[1,:]-q[0,:]) # 1st
#    dqdym[1,:]=(q[2,:]-q[0,:])/2. # 2nd
    dqdym[2,:]=(2*q[3,:]+3*q[2,:]-6*q[1,:]+q[0,:])/6. # 3rd
    dqdym[-2,:]=(2*q[-1,:]+3*q[-2,:]-6*q[-3,:]+q[-4,:])/6. # 3rd
    dqdym[3:-2,:]=(-3*q[5:,:]+30*q[4:-1,:]+20*q[3:-2,:]-60*q[2:-3,:]+15*q[1:-4,:]-2*q[:-5,:])/60. # 5th

    dqdyp[-2,:]=(q[-1,:]-q[-2,:]) # 1st
 #   dqdyp[-2,:]=(q[-1,:]-q[-3,:])/2. # 2nd
    dqdyp[1,:]=-(2*q[0,:]+3*q[1,:]-6*q[2,:]+q[3,:])/6. # 3rd
    dqdyp[-3,:]=-(2*q[-4,:]+3*q[-3,:]-6*q[-2,:]+q[-1,:])/6. # 3rd
    dqdyp[2:-3,:]=-(-3*q[:-5,:]+30*q[1:-4,:]+20*q[2:-3,:]-60*q[3:-2,:]+15*q[4:-1,:]-2*q[5:,:])/60. # 5th

    dqdx=np.where(u>0.,dqdxm,dqdxp)
    dqdy=np.where(v>0.,dqdym,dqdyp)

    dqdt+=-u*dqdx*rdxa
    dqdt+=-v*dqdy*rdya
    return dqdt

###############################
# for Cartesian diffusion
def diffuse_cart(q,dcoefx,dcoefy,grid):
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    rdxc=grid['rdxc']
    rdyc=grid['rdyc']
    dqdt=0.*q
    fqx=-dcoefx*(q[:,1:]-q[:,:-1])*rdxc
    fqy=-dcoefy*(q[1:,:]-q[:-1,:])*rdyc
    dqdt[:,1:-1] += (-fqx[:,1:] + fqx[:,:-1])*rdxa[:,1:-1] 
    dqdt[:,0] += -2.*fqx[:,1]*rdxa[:,0] 
    dqdt[:,-1] += 2.*fqx[:,-1]*rdxa[:,-1] 
    dqdt[1:-1,:] += (-fqy[1:,:] + fqy[:-1,:])*rdya[1:-1,:] 
    dqdt[0,:] += -2.*fqy[1,:]*rdya[0,:] 
    dqdt[-1,:] += 2.*fqy[-1,:]*rdya[-1,:] 
    return dqdt

################################
# for axisymmetric diffusion of w and scalars
def diffuse_axi_w(q,dcoefx,dcoefy,grid):
    rc=grid['xc']
    rdrr=grid['rdrr']
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    rdxc=grid['rdxc']
    rdyc=grid['rdyc']
    dqdt=0.*q
    fqx=-dcoefx*rc*(q[:,1:]-q[:,:-1])*rdxc
    fqy=-dcoefy*(q[1:,:]-q[:-1,:])*rdyc
    dqdt[:,1:-1] += (-fqx[:,1:] + fqx[:,:-1])*rdrr[:,1:-1] 
    dqdt[:,0] += -fqx[:,0]*rdrr[:,0] 
    dqdt[:,-1] += fqx[:,-1]*rdrr[:,-1] 
    dqdt[1:-1,:] += (-fqy[1:,:] + fqy[:-1,:])*rdya[1:-1,:] 
    dqdt[0,:] += -2.*fqy[1,:]*rdya[0,:] 
    dqdt[-1,:] += 2.*fqy[-1,:]*rdya[-1,:] 
    return dqdt

def diffuse_axi_v(q,dcoefx,dcoefy,grid,semic=0.): 
# use d/dr(1/r d/dr(r v) ) = 1/r^2 d/dr ( r^3 d/dr(v/r))
# and exploit that v/r is constant in solid body rotation near axis
# additional axisymmetric diffusion for radial and azimuthal
# semi-slip coefficient semic (a.k.a. gamma) is applied 
# warning: make sure r3 is for pressure grid and r2 for u grid
    r=grid['x']
    r2=grid['xa2']
    r3=grid['xc3']
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    rdxc=grid['rdxc']
    rdyc=grid['rdyc']
    dqdt=0.*q
    qor=0*q
    qor[:,1:]=q[:,1:]/r[:,1:]  # when q=v, this will be angular velocity
    fqx=-dcoefx*(qor[:,1:]-qor[:,:-1])*rdxc*r3
    fqy=-dcoefy*(q[1:,:]-q[:-1,:])*rdyc
    dqdt[:,1:-1] += (-fqx[:,1:] + fqx[:,:-1])*rdxa[:,1:-1]/r2[:,1:-1] 
#    dqdt[:,0] += 0. # seems strange, but note v and u remain 0 at r=0 
    dqdt[:,-1] += 2.*fqx[:,-1]*rdxa[:,-1]/r2[:,-1] 
    dqdt[1:-1,:] += (-fqy[1:,:] + fqy[:-1,:])*rdya[1:-1,:] 
#    dqdt[0,:] += -2.*fqy[1,:]*rdya[0,:] # free-slip  
    dqdt[0,:] += -2.*( fqy[1,:] + semic*q[0.,:] )*rdya[0,:] # semi-slip added 30 May 2017 
    dqdt[-1,:] += 2.*fqy[-1,:]*rdya[-1,:] 
    return dqdt


# divergence of u and v, on the pressure grid
def div2(u,v,grid,axisym=False):
    rdxc=grid['rdxc']
    rdyc=grid['rdyc']
    r=grid['x']
    rdrra=grid['rdrra'] 
    if axisym:
        ur=u*r
        return .5*( ur[:-1,1:] + ur[1:,1:] - ur[:-1,:-1] - ur[1:,:-1])*rdrra[0:-1,:] + \
           .5*( v[1:,:-1] + v[1:,1:] - v[:-1,:-1] - v[:-1,1:])*rdyc[:,0:-1]
    else:
        return .5*( u[:-1,1:] + u[1:,1:] - u[:-1,:-1] - u[1:,:-1])*rdxc[0:-1,:] + \
               .5*( v[1:,:-1] + v[1:,1:] - v[:-1,:-1] - v[:-1,1:])*rdyc[:,0:-1]
###################
# pressure gradient, note p is staggered from (u,v) grid
def pgf(p,rdx,rdy):
    dudt=np.zeros(rdx.shape)
    dvdt=np.zeros(rdx.shape)
    dpdy=( p[1:,:]-p[:-1,:] )
    dvdt[1:-1,1:-1] = -.5* ( dpdy[:,:-1] + dpdy[:,1:] ) * rdy[1:-1,1:-1]
    dvdt[1:-1,0] =  -dpdy[:,0] * rdy[1:-1,0]
    dvdt[1:-1,-1] = -dpdy[:,-1] * rdy[1:-1,-1]
    dpdx=( p[:,1:]-p[:,:-1] )
    dudt[1:-1,1:-1] = -.5* ( dpdx[:-1,:] + dpdx[1:,:] ) * rdx[1:-1,1:-1]
    dudt[0,1:-1] =  -dpdx[0,:] * rdx[0,1:-1]
    dudt[-1,1:-1] = -dpdx[-1,:] * rdx[-1,1:-1]
    return dudt,dvdt

#########################
# iterative solver to find pressure that makes
# the velocity field incompressible
def incomp3(u,v,p,divtorid,grid,limcoef=None,pit=40,alpha=.5,
            noslip=[False,False,False,False],axisym=False):
    rdxa=grid['rdxa']
    rdya=grid['rdya']
    npass=0
    if limcoef==None: limcoef=1./(rdxa[:-1,:-1]**2+rdya[:-1,:-1]**2)
    pa=p.copy()
    while npass<pit:
        pgx,pgy=pgf(pa,rdxa,rdya)
        r=u+pgx
        s=v+pgy
        velbc(r,s,noslip)
        dv=div2(r,s,grid,axisym=axisym)
        pa = pa - alpha*limcoef*(dv+divtorid)
        npass+=1
    return r,s,pa

##############################
# Adams-Bashforth coefficients for variable dt     
def abcoef(p,q):
# assumes you have used dt, dt1 and dt2 (current to past)
# to calculate p = dt1/dt, q=(dt1+dt2)/dt
    ab=np.zeros((3,3))
# the 1st order coefficents:
    ab[0,2]=0.
    ab[0,1]=0.
    ab[0,0]=1.
# the 2nd order coefficents:
    ab[1,2]=0.
    ab[1,1]=-0.5/p
    ab[1,0]=1-ab[1,1]
# the 3rd order coefficients:
    ab[2,2]=(2+3*p)/(6*q*(q-p))
    ab[2,1]=-(1+2*q*ab[2,2])/(2*p)
    ab[2,0]=1.-ab[2,1]-ab[2,2]
    return ab

##########################
# updates q given three levels of dq/dt
def abstep(dqdta,ab):
    if dqdta[2] is not None:
        return ab[0]*dqdta[0]+ab[1]*dqdta[1]+ab[2]*dqdta[2]
    elif dqdta[1] is not None:
        return ab[0]*dqdta[0]+ab[1]*dqdta[1]
    else:
        return ab[0]*dqdta[0]

#################################
# used for the extreme of semi-slip: no-slip
def velbc(u,v,noslip=[False,False,False,False]):
    if noslip[0]: u[0,:]=0. 
    if noslip[1]: v[:,-1]=0. 
    if noslip[2]: u[-1,:]=0. 
    if noslip[3]: v[:,0]=0. 
    return

#################################
# used to set azimuthal velocity to no-slip
def velbca(a,noslip=[False,False,False,False]):
    if noslip[0]: a[0,:]=0. 
    if noslip[1]: a[:,-1]=0. 
    if noslip[2]: a[-1,:]=0. 
    if noslip[3]: a[:,0]=0. 
    return

###################################
# interpolates pressure p to the velocity grid
# (used only for plotting and grid doubling)
def p2pu(p):
    pu=np.zeros((p.shape[0]+1,p.shape[1]+1))
    pu[1:-1,1:-1] = .25*( p[:-1,:-1]+p[1:,:-1]+p[1:,1:]+p[:-1:,1:] )
    pu[1:-1,0] = .5* (p[:-1,0] + p[1:,0])
    pu[1:-1,-1] = .5* (p[:-1,-1] + p[1:,-1])
    pu[0,1:-1] = .5* (p[0,:-1] + p[0,1:])
    pu[-1,1:-1] = .5* (p[-1,:-1] + p[-1,1:])
    pu[0,0]=p[0,0]
    pu[0,-1]=p[0,-1]
    pu[-1,0]=p[-1,0]
    pu[-1,-1]=p[-1,-1]
    return pu

##################################
# tests the makexy and makegridcoef functions
def npcfdtest():
    x,y=makexy(6,4,2.,1.,.5,1.)
    g=makegridcoef(x,y)
    print(g['x'])
    print(g['y'])
    print( g.keys() )
if __name__=='__main__':
	npcfdtest()
